/*****************************************************************************/
/*                                                                           */
/* You may copy, modify, and distribute this sample program in any           */
/* form without payment to IBM,  for any purpose including developing,       */
/* using, marketing or distributing programs that include or are             */
/* derivative works of the sample program.                                   */
/*                                                                           */
/* The sample program is provided to you on an "AS IS" basis, without        */
/* warranty of any kind.  IBM HEREBY EXPRESSLY DISCLAIMS ALL WARRANTIES,     */
/* EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       */
/* Some jurisdictions do not allow for the exclusion or limitation of        */
/* implied warranties, so the above limitations or exclusions may not        */
/* apply to you.  IBM shall not be liable for any damages you suffer as      */
/* a result of using, modifying or distributing the sample program or        */
/* its derivatives.                                                          */
/*																			 */
/*  * @author Shravan														 */
/*                                                                           */
/*                                                                           */
/* This is the core/main file of MQSI broker side operations.		     	 */
/*      								     								 */
/* V 1.0 - Basic features						     						 */
/* V 1.1 - XMLNS support						     						 */
/* V 1.2 - MRM support						             					 */	
/* V 1.3 - Bug fix 'firstParserClassName' to be set in the node constructor  */
/* V 1.4 - Support for OS Logging					     					 */
/* V 1.6 - Multiple instances support.										 */
/* V 1.8 - Message transaction support.										 */
/* V 1.9 - Common class to hold node attributes values created				 */
/* V 2.0 - Change wrt OsLogData first parameter								 */
/* ************************************************************************* */

package com.ibm.Shra.Xternal.Plugins;

import com.ibm.broker.plugin.MbMessage;
import com.ibm.broker.plugin.MbException;
import com.ibm.broker.plugin.MbBrokerException;
import com.ibm.broker.plugin.MbRecoverableException;
import com.ibm.broker.plugin.MbConfigurationException;
import com.ibm.broker.plugin.MbInputNode;
import com.ibm.broker.plugin.MbInputNodeInterface;
import com.ibm.broker.plugin.MbOutputTerminal;
import com.ibm.broker.plugin.MbMessageAssembly;
import com.ibm.broker.plugin.MbService;
import com.ibm.broker.plugin.MbBroker;


public class AnyQmgrInputNode extends MbInputNode implements MbInputNodeInterface
{
	private String className = "AnyQmgrInputNode";
	/*
	 * This input node has two terminals output and failure...
	 */
	private MbOutputTerminal outTerminal = null;
	private MbOutputTerminal failureTerminal = null;
	private static final String OUTPUT_TERMINAL = "out"; 
	private static final String FAILURE_TERMINAL = "failure";

	private logData logdt = null; /*to log the information into a file*/
	private MQOps mqOperations = null; /* object to do QMgr operations*/
	/* All node attributes are now defined in NodeAttributes.java*/
	private boolean initialized = false; /*to check whether node is intialized or not*/
	private boolean queueFlag = true; /*to indicate/check that queue names are specified on the node properties*/	
	private boolean msgState = true; /*to indicate/check whether the message is to be committed or rolledback*/

	public AnyQmgrInputNode() throws MbException
	{
		super();

		setOutTerminal(createOutputTerminal(OUTPUT_TERMINAL));
		setFailureTerminal(createOutputTerminal(FAILURE_TERMINAL));

		/*
		 * Donot set this to MQHMD if your message doesnot contian the MQMD header...
		 */			
		this.setAttribute("firstParserClassName", "MQHMD");
		
		logdt = logData.getInstance();
		/*logdt.setlogDataLevel("DEBUG"); //Only used while testing..:-)*/
		mqOperations = new MQOps();			
	}

	public int run(MbMessageAssembly assembly) throws MbException
	{	
		String methodName = "run";
		byte[] msgIn = null; /*this is the message received from Queue*/
		MbMessageAssembly newMsgAssembly = null;
		try
		{
			if (!initialized)
				{
					logdt.setlogDataLevel(getLogLevel()); /*get the logLevel from the node properties*/					
					initializeAll();
					OSLogData.syslog(className,methodName,OSLogData.SYSLOG_INFO,"NODE_INITIALIZED", null);				
				}
			logdt.logData_Debug(" { "+className+"."+methodName);	
			if(!queueFlag)
		    		return FAILURE_RETURN;			

			if(mqOperations != null)
			{
				msgIn = mqOperations.getMessage(); /*You get the message here in bytes along with MQMD header*/

				if (msgIn == null)
				{
					/*.
					 * ..no message available on any queue or u reached the last Q...so sleep for a while (if necessary)
					 * and continue....We should handle the control to the broker instead of waiting indefinitely on the queue.
					 */
					Thread.sleep(5000);	
					logdt.logData_Debug(" } "+className+"."+methodName);
					return TIMEOUT;
				}
				else
				{
					/*
					 * Valid message is read form the queue so create MbMessage and Assembly anf transfer it to the output node
					 */
					logdt.logData_Info("  Message Received from Queue is : "+msgIn);
					logdt.logData_Debug("  Creating MbMessage ");
					msgState = true; /*By default assume that we are going to commit the message*/
					
					MbMessage msg = createMessage(msgIn);
					msg.finalizeMessage(MbMessage.FINALIZE_VALIDATE);

					newMsgAssembly = new MbMessageAssembly(assembly,msg);
					/*logdt.logData_Debug("  MessageAssembly created "+newMsgAssembly.toString());*/
					logdt.logData_Debug("  Message propagating to output terminal ");
		
					/* 
					 * Now propagate the message assembly. If the terminal is not attached, an exception will be thrown.
					 * The user can choose to handle this exception, but it is not neccessary since the framework
					 * will catch it and propagate the message to the failure terminal.
					*/
					dispatchThread();			
					if(getOutTerminal().isAttached())
					{
						getOutTerminal().propagate(newMsgAssembly);
						logdt.logData_Info("  Message propagated to output terminal successfully");
	/*					mqOperations.acknowledgeMessage();//We will acknowledge the msg in the finally block*/
						msgIn = null;
					}
					else
					{
						/*Infact this should be a FATAL ERROR ....and should handle the way built-in input node
						 * does, anyhow the frame work does that..but for information just log it.*/
						logdt.logData_Error("  ERROR ... OUT TERMINAL NOT CONNECTED :-( ");
						logdt.logData_Debug(" } "+className+"."+methodName);
						msgState = false; /*Uncommit the message that is picked up*/
						return FAILURE_CONTINUE;
						//return FAILURE_RETURN;
					}
				}
			 }
		}
		catch(InterruptedException ie) 
		{
			msgState = false; /*Uncommit/rollback the message that is picked up*/
		 	logdt.logData_Error("\tInterrupted Exception occurred ",ie);	
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"INTERNAL_UNKNOWN_ERROR", logdt.getExceptionString(ie));
			/* 	throw ie;*/
		}
		/* 
		 * Catch the MbException and log it to the file and rethrows the original exception.
		 */
		catch(MbException mbex)
		{
			msgState = false; /*Uncommit/rollback the message that is picked up*/
			logdt.logData_Error("\tMbException occurred ",mbex);
			parseBrokerException(mbex, 0);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"WBIMB_ERROR",logdt.getExceptionString(mbex));
			throw mbex;
		}
		/* 
		 * Catch the MbBrokerException and log it to the file and rethrows the original exception.
		 */
		catch(MbBrokerException mbbrke)
		{
			msgState = false; /*Uncommit/rollback the message that is picked up*/
			logdt.logData_Error("\tMbBrokerException occurred ",mbbrke);			
			/*Internal error occurred while playing with the message, We don't have to handle this so push it to broker and 
			 * it will take care of sending the msg to "failure" terminal...
			 */
			//if (failureTerminal.isAttached())
			//{
			//	mqOperations.acknowledgeMessage();
			//	failureTerminal.propagate(newMsgAssembly);
			//	logdt.logData_Info("  Message propagated to failure terminal ");
			//	logdt.logData_Debug(" } AnyQmgrInputNode::run ");
			//	return FAILURE_RETURN;
			//}
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"WBIMB_ERROR",logdt.getExceptionString(mbbrke));
			throw mbbrke;	
		}
		catch(Exception ex)
		{
			msgState = false; /*Uncommit/rollback the message that is picked up*/
			logdt.logData_Error("\tException occurred ",ex);			
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_WARN,"INTERNAL_UNKNOWN_WARN",logdt.getExceptionString(ex));
		}
		finally
		{
			if(newMsgAssembly != null)
			{
				if(msgState)/*Decide to acknowledge/rollback the message that is picked up*/
				{
					logdt.logData_Debug("  Message acknowledged or Committed.");
					mqOperations.acknowledgeMessage();
				}
				else
				{
					logdt.logData_Debug("  Message backedout or Uncommitted. ");
					mqOperations.backoutMessage();
				}
				logdt.logData_Debug("  Destroying the message Assembly ");			
				destroyMessageAssembly(newMsgAssembly);	
			}
		}
		logdt.logData_Debug(" } "+className+"."+methodName);
		if(msgState)
				return SUCCESS_RETURN;
		else 		return SUCCESS_CONTINUE;
	}
	
	/*
	 * This function need to be called only once as part of initializing the node...
	 * */	
	private synchronized void initializeAll() throws MbException,MbRecoverableException
	{
		String methodName = "initializeAll";
		try
		{
		Thread.currentThread().setName(this.getBroker().getName()+"."+this.getExecutionGroup().getName()+"."+this.getMessageFlow().getName());
			logdt.logData_Debug(" { "+className+"."+methodName);
				logdt.logData_Debug("  firstParserClassName is set to "+this.getAttribute("firstParserClassName"));
				/*
				 * if you want your messages to be of XML or XMLNS or MRM, then set it over here instead of BLOB
				 */
				this.setAttribute("messageDomainProperty", getDomain());/*Default is BLOB*/
				/*
				 * Please read the documentaion clearly before setting these options. If you use 'MRM' then you have 
				 * to set certain other properties too....
				 */
				logdt.logData_Info("  Setting messageDomainProperty to "+this.getAttribute("messageDomainProperty")); 
				if(getDomain().equals("MRM")) /*Time to set the details and all are mandatory*/
				{
					if (getMessageSet() != null)
						{	
							this.setAttribute("messageSetProperty", getMessageSet());/* Set the Message Set */
							logdt.logData_Debug("  messageSetProperty is set to "+this.getAttribute("messageSetProperty"));
						}
					else 
						{
							logdt.logData_Debug("  messageSetProperty is not set as no value is given.");
							throw new MbConfigurationException(className,methodName,null,null," MESSAGE SET PROPERTY CANNOT BE EMPTY AS DOMAIN IS SET TO MRM",null);
						}
					if (getMessageType() != null)
						{	
							this.setAttribute("messageTypeProperty", getMessageType());/* Set the Message Type */
							logdt.logData_Debug("  messageTypeProperty is set to "+this.getAttribute("messageTypeProperty"));
						}
					else 
						{
							logdt.logData_Debug("  messageSetType is not set as no value is given.");
							throw new MbConfigurationException(className,methodName,null,null," MESSAGE TYPE PROPERTY CANNOT BE EMPTY AS DOMAIN IS SET TO MRM",null);
						}
					if (getMessageFormat() != null)
						{	
							this.setAttribute("messageFormatProperty", getMessageFormat());/* Set the Message Format */
							logdt.logData_Debug("  messageFormatProperty is set to "+this.getAttribute("messageFormatProperty"));
						}
					else 
						{
							logdt.logData_Debug("  messageSetFormat is not set to as no value is given.");
							throw new MbConfigurationException(className,methodName,null,null," MESSAGESET PROPERTY CANNOT BE EMPTY AS DOMAIN IS SET TO MRM",null);
						}
				}

			if(NodeAttributes.queues.equals(""))
			{
				logdt.logData_Error("  ERROR : QUEUE NAME(s) EMPTY ");
				/*This condition has to be a FATAL Exception and should stop the broker in continueing the process of 				initalization of the node...*/
				initialized = false; /*do this so that initialization of node is looped and message processing is not done*/
				queueFlag = false; /*just specify the node to terminate as there are no queue name(s) in it*/
				throw new MbRecoverableException(className,methodName,null,null," QUEUE NAME(s) CANNOT BE EMPTY ",null);
				/*throw new MbConfigurationException("AnyQmgrInputNode.java","AnyQmgrInputNode::initializeAll()",null,null,"QUEUE NAME(s) CANNOT BE EMPTY",null);*/
			}
			else
			{	
				mqOperations.init();		
			}
			initialized = true; /*do this as initialization of node has to be done only once...*/
		}catch(MbRecoverableException mbrex)
		{
			logdt.logData_Error("\tMbRecoverableException Detailed Stack Trace : ",mbrex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"WBIMB_ERROR", logdt.getExceptionString(mbrex));			
			throw mbrex;
		}catch(MbException mbex)
		{
			logdt.logData_Error("\tMbException Detailed Stack Trace : ",mbex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"WBIMB_ERROR",logdt.getExceptionString(mbex));			
			parseBrokerException(mbex, 0);
			throw mbex;
		}

		/*catch(Exception ex) 
		{
			logdt.logData_Error("\tIn Excepion ",ex);
		}*/
       	 logdt.logData_Debug(" } "+className+"."+methodName);
	}

/*
 * Set the Output Terminal
 * */
    void setOutTerminal(MbOutputTerminal outTerm)
    {
    	outTerminal = outTerm;	
    }
    MbOutputTerminal getOutTerminal()
    {
    	return outTerminal;
    }
    
    /*
 * Set the Failure Terminal
 * */
    void setFailureTerminal(MbOutputTerminal failTerm)
    {
    	failureTerminal = failTerm;	
    }
    MbOutputTerminal getFailureTerminal()
    {
    	return failureTerminal;
    }

	/*for node properties attribute 'queueManager'*/
	public String getQueueManager() 
	{
		/*If no queueManager is specified then use broker queueManger as the queueManager...*/
		try
		{ 
			if(NodeAttributes.queueManager == null)
		  		NodeAttributes.queueManager = getBroker().getQueueManagerName();
		}catch(MbException mbe)
		{
			logdt.logData_Error("\tIn MbException ",mbe);
//			OSLogData.syslog(this, "AnyQmgrInputNode::getQueueManager",OSLogData.SYSLOG_WARN,"WBIMB_WARN", logdt.getExceptioString(mbe));			
		}
		logdt.logData_Debug("  "+className+".getQueueManager returns "+NodeAttributes.queueManager);
		return NodeAttributes.queueManager;
	}
	
	/*for node properties attribute 'queueManager' used by broker*/
	public void setQueueManager(String Qmgr)
	{
		
		try
		{
			//If no queueManager is specified then use broker queueManger as the queueManager
			if(Qmgr == null)
					NodeAttributes.queueManager = getBroker().getQueueManagerName();
			else 	NodeAttributes.queueManager = Qmgr;
		}catch(MbException mbe)
		{
			logdt.logData_Error("\tIn MbException ",mbe);
		}
		
		logdt.logData_Debug("  "+className+".setQueueManager sets "+NodeAttributes.queueManager);
	}
	
	/*for node properties attribute 'hostName'*/
	public String getHost()
	{
		logdt.logData_Debug("  "+className+".getHost returns "+NodeAttributes.hostName);
		return NodeAttributes.hostName;
	}
	
	/*for node properties attribute 'hostName' used by broker*/
	public void setHost(String hst)
	{
		/*if local this can blank*/
		if (hst != null)
			NodeAttributes.hostName = hst;
		logdt.logData_Debug("  "+className+".setHost sets "+NodeAttributes.hostName);
	}
	
	/*for node properties attribute 'port'*/
	public String getPort()
	{
		logdt.logData_Debug("  "+className+".getPort returns "+Integer.toString(NodeAttributes.port));
		return Integer.toString(NodeAttributes.port);
	}
	
	/*for node properties attribute 'port' used by broker*/
	public void setPort(String prt)
	{
		try
		{
			if(prt == null) 
				return;
			NodeAttributes.port = Integer.parseInt(prt);
		}catch(NumberFormatException nfe){
		OSLogData.syslog(className,"setPort",OSLogData.SYSLOG_WARN,"WBIMB_WARN",logdt.getExceptionString(nfe));			
		}
		catch(Exception e){
		OSLogData.syslog(className,"setPort",OSLogData.SYSLOG_WARN,"WBIMB_WARN",logdt.getExceptionString(e));			
		}		
		logdt.logData_Debug("  "+className+".setPort sets "+NodeAttributes.port);
	}
	
	/*for node properties attribute 'queues'*/
	public String getQueues()
	{
		logdt.logData_Debug("  "+className+".getQueues returns "+NodeAttributes.queues);
		return NodeAttributes.queues;
	}
	
	/*for node properties attribute 'queues'used by broker*/
	public void setQueues(String que)
	{ 
	    /*use string tokenizer to divide the list of queues*/
		NodeAttributes.queues = que;
		logdt.logData_Debug("  "+className+".setQueues sets "+NodeAttributes.queues);
	}
	
	/*for node properties attribute 'channel'*/
	public String getChannel()
	{
		logdt.logData_Debug("  "+className+".getChannel returns "+NodeAttributes.channel);
		return NodeAttributes.channel;
	}
	/*for node properties attribute 'channel' used by broker*/
	public void setChannel(String chl)
	{
		NodeAttributes.channel = chl;
		logdt.logData_Debug("  "+className+".setChannel sets "+NodeAttributes.channel);
	}

	/*for node properties attribute 'domain'*/
	public String getDomain()
	{
		logdt.logData_Debug("  "+className+".getDomain returns "+NodeAttributes.domain);
		return NodeAttributes.domain;
	}
	/*for node properties attribute 'domain' used by broker*/
	public void setDomain(String dom)
	{
		NodeAttributes.domain = dom;
		logdt.logData_Debug("  "+className+".setDomain sets "+NodeAttributes.domain);
	}

	/*for node properties attribute 'messageSet'*/
	public String getMessageSet()
	{
		logdt.logData_Debug("  "+className+".getMessageSet returns "+NodeAttributes.messageSet);
		return NodeAttributes.messageSet;
	}
	/*for node properties attribute 'messageSet' used by broker*/
	public void setMessageSet(String mset)
	{
		NodeAttributes.messageSet = mset;
		logdt.logData_Debug("  "+className+".setMessageSet sets "+NodeAttributes.messageSet);
	}
	/*for node properties attribute 'messageType'*/
	public String getMessageType()
	{
		logdt.logData_Debug("  "+className+".getMessageType returns "+NodeAttributes.messageType);
		return NodeAttributes.messageType;
	}
	/*for node properties attribute 'messageType' used by broker*/
	public void setMessageType(String mtype)
	{
		NodeAttributes.messageType = mtype;
		logdt.logData_Debug("  "+className+".setMessageType sets "+NodeAttributes.messageType);
	}
	/*for node properties attribute 'messageFormat'*/
	public String getMessageFormat()
	{
		logdt.logData_Debug("  "+className+".getMessageFormat returns "+NodeAttributes.messageFormat);
		return NodeAttributes.messageFormat;
	}
	/*for node properties attribute 'messageFormat' used by broker*/
	public void setMessageFormat(String mformat)
	{
		NodeAttributes.messageFormat = mformat;
		logdt.logData_Debug("  "+className+".setMessageFormat sets "+NodeAttributes.messageFormat);
	}	
	/*for node properties attribute 'transaction'*/
	public String getTransaction() 
	{
		logdt.logData_Debug("  "+className+".getTransaction returns "+NodeAttributes.transaction);
		return NodeAttributes.transaction;
	}

	/*for node properties attribute 'transaction' used by broker*/
	public void setTransaction(String transaction) 
	{
		NodeAttributes.transaction = transaction;
		logdt.logData_Debug("  "+className+".setTransaction sets "+transaction);
	}

	/*for node properties attribute 'logLevel'*/
	public String getLogLevel()
	{
		String level = null;
		
		switch (NodeAttributes.logLevel)
		{
			case logData.LOGDATA_LEVEL_NONE: 
										level = "NONE";
										break;
			case logData.LOGDATA_LEVEL_DEBUG:
										level = "DEBUG";
										break;
			case logData.LOGDATA_LEVEL_INFO:
			 							level = "INFO";
										break;
			case logData.LOGDATA_LEVEL_ERROR:
										level = "ERROR";
										break;
		}		
		logdt.logData_Debug("  "+className+".getLogLevel returns "+level);
		return level;	
	}
	 
	/*for node properties attribute 'logLevel' used by broker*/
	public void setLogLevel(String level)
	{
		if(level == null) return;
		
		if(level.equals("NONE"))
			NodeAttributes.logLevel = logData.LOGDATA_LEVEL_NONE;
		else if(level.equals("DEBUG")) 
			NodeAttributes.logLevel = logData.LOGDATA_LEVEL_DEBUG;
		else if(level.equals("INFO")) 
			NodeAttributes.logLevel = logData.LOGDATA_LEVEL_INFO;
		else if(level.equals("ERROR"))
			NodeAttributes.logLevel = logData.LOGDATA_LEVEL_ERROR;
		
		logdt.logData_Debug("  "+className+".setLogLevel sets "+level);
	}
	

	/*This is essential to write though not mandatory. If this is not set then Garbage Collector would clean
	the memroy, but if your message throughput is too high then its not predictable when GC would be called and
	in no time system resources would be drained. So its a good practise to clean the message explicitly.*/	
	private void destroyMessageAssembly(MbMessageAssembly assembly)
	{
		logdt.logData_Debug(" { "+className+".destroyMessageAssembly ");
		try
		{
			if( assembly.getMessage() != null)//checck for if message is enpty or not first
				assembly.getMessage().clearMessage();
		}
		catch(MbException mbe)
		{
			logdt.logData_Error("  MbException occurred ",mbe);
			OSLogData.syslog(className,"destroyMessageAssembly",OSLogData.SYSLOG_WARN,"WBIMB_WARN",logdt.getExceptionString(mbe));			
		}
		logdt.logData_Debug(" } "+className+".destroyMessageAssembly ");
	}
	
	public static String getNodeName()
	{
		return "AnyQmgrInputNode";
	}

	/*This method is called by the broker when stop command is issued on the flow by Execution Group.
	Though its not mandatory to implement this method but its a good practise to code this function...
	mostly this is very helpful to clean the allocated resources by the node...like the way its done here...*/	
	public void onDelete()
	{
		logdt.logData_Debug(" { "+className+".onDelete ");
		if (initialized)
			mqOperations.releaseResources();
		logdt.closeInstance();	
		logdt.logData_Debug(" } "+className+".onDelete ");
		initialized = false;
		OSLogData.syslog(className,"onDelete",OSLogData.SYSLOG_INFO,"NODE_DELETED", null);				
	}
	private void parseBrokerException(MbException mbex, int pLevel)
	{
	    if(mbex != null)
	      {
		logdt.logData_Error("\t parseStackErrorLevel : "+pLevel);
		logdt.logData_Error("\t Exception Trace Text : "+ mbex.getTraceText());

	        /* Now traverse the hierarchy */
        	MbException e[] = mbex.getNestedExceptions();
	        int size = e.length;
        	for(int i = 0; i < size; i++)
	          {
        	    parseBrokerException(e[i], pLevel + 1);
	          }
	      }
	}

}