/*****************************************************************************/
/*                                                                           */
/* You may copy, modify, and distribute this sample program in any           */
/* form without payment to IBM,  for any purpose including developing,       */
/* using, marketing or distributing programs that include or are             */
/* derivative works of the sample program.                                   */
/*                                                                           */
/* The sample program is provided to you on an "AS IS" basis, without        */
/* warranty of any kind.  IBM HEREBY EXPRESSLY DISCLAIMS ALL WARRANTIES,     */
/* EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       */
/* Some jurisdictions do not allow for the exclusion or limitation of        */
/* implied warranties, so the above limitations or exclusions may not        */
/* apply to you.  IBM shall not be liable for any damages you suffer as      */
/* a result of using, modifying or distributing the sample program or        */
/* its derivatives.                                                          */
/*																			 */
/*  * @author Shravan														 */
/*                                                                           */
/*                                                                           */
/*This is the core/main file of MQSeries queue manager side operations.      */
/* V 1.0 - Basic features												     */
/* V 1.4 - Support for OS Logging					     					 */
/* V 1.9 - Code pertaining to the node attributes values access, is deleted  */
/* 			as new class NodeAttributes.java created.						 */
/* V 2.0 - Change wrt OsLogData first parameter								 */
/* ************************************************************************* */
package com.ibm.Shra.Xternal.Plugins;

import com.ibm.mq.MQEnvironment;
import com.ibm.mq.MQException;
import com.ibm.mq.MQQueueManager;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import com.ibm.broker.plugin.MbFatalException;

public class MQOps
{
	private String className = "MQOps";
	private static logData ld =  null;
	private MQQueueManager qMgr = null; /*Object to connect to the queue Manager*/
	
	private QueueOps[] qopsList = null; /*List of MQ Queues*/
	private int qopsListSize; /*Total number of MQ Queues*/
	private int qopsListIndex = 0; /*This will always point to the current Queue to get the message.*/
	private String deadLetterQueueName = null;
	private int qmgr_CCSID ;
	
	public MQOps()
	{
		super();
		ld = logData.getInstance();
		ld.logData_Debug(" { "+className);
		MQException.log = null; /*Disabling this causes errors not to be written to System.err()..as i want every 
		error message pertaining to the node to be logged into the file*/
		ld.logData_Debug(" } "+className);
	}

	public void init()
	{
		String methodName = "init";
		ld.logData_Debug(" { "+className+"."+methodName);
		try
		{
			if(NodeAttributes.hostName != null)
				{
					MQEnvironment.hostname = NodeAttributes.hostName; /* set the hostname to connect to ..applicable only if your 
					queue Manager is remote/notLocal*/
					ld.logData_Debug("  Host name set as : "+NodeAttributes.hostName);
				}
			if(NodeAttributes.port != -1)
				{
					MQEnvironment.port = NodeAttributes.port; /* set the port number to listen at...applicable only if your
					queue Manager is remote*/
					ld.logData_Debug("  Port number set as : "+NodeAttributes.port);
				}
			if(NodeAttributes.channel != null)
				{
					MQEnvironment.channel = NodeAttributes.channel; /* set the CASE-SENSITIVE name of the SVR CONN channel 
					on the remote queue manager*/
					ld.logData_Debug("  Channel name set as : "+NodeAttributes.channel);
				}
			if(NodeAttributes.queueManager != null) 
				{	
					/*When we reach here queueManager name should NOT be null...it can be name received from node 
					property attribute or should use broker queuemanager*/
					//MQEnvironment.properties.put((Object)"MQC.TRANSPORT_PROPERTY",(Object)"MQC.TRANSPORT_MQSERIES");
					MQEnvironment.properties.put("MQC.TRANSPORT_PROPERTY","MQC.TRANSPORT_MQSERIES");
					qMgr = new MQQueueManager(NodeAttributes.queueManager);
					/* Inquire to find the DeadLetterQueue*/
					
					 int[]  selectors = { MqConstants.MQIA_CODED_CHAR_SET_ID, MqConstants.MQCA_DEAD_LETTER_Q_NAME	};
					 int[]  intAttrs = new int[1];
					 byte[] charAttrs = new byte[MqConstants.MQ_Q_NAME_LENGTH];
					 
					qMgr.inquire(selectors,intAttrs,charAttrs);
					qmgr_CCSID = intAttrs[0];
					deadLetterQueueName = new String(charAttrs);
					deadLetterQueueName=deadLetterQueueName.trim();
				 	 if(deadLetterQueueName.equals(""))
				 	 	deadLetterQueueName = new String("SYSTEM.DEAD.LETTER.QUEUE"); //Set it to default DLQ.
					ld.logData_Debug("  QMgr CCSID is "+qmgr_CCSID);
  	   		  		ld.logData_Debug("  QMgr DeadLetterQueue name is "+deadLetterQueueName);
					ld.logData_Debug("  Connected to QueueManager : "+NodeAttributes.queueManager);
			}
								
			if( NodeAttributes.queues != null)
				{ /* By now queue(s) are recieved as one single string with comma separated between queue names*/
					ld.logData_Debug("  Establish / Parse all queue names ");
					parseQueueList(); /* Identify the queue name by parsing them and create a list of queues*/
					ld.logData_Debug("  Establish Connection To All Queues ");
					for(qopsListIndex=0;qopsListIndex<qopsListSize;qopsListIndex++)
					{
						qopsList[qopsListIndex].getConnected(qMgr); /*Connect all queues*/
					}
					ld.logData_Debug("  Connection Established To All Queues ");
				}
		}
		catch (MQException ex)
		{
			ld.logData_Error("\tWMQ Exception Occurred, Completion code " +ex.completionCode + " Reason code " + ex.reasonCode);
			ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",ex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR", ld.getExceptionString(ex));
		}
		qopsListIndex=0; /*this is essential...point to the first queue*/
		ld.logData_Debug(" } "+className+"."+methodName);
	}
	
	/*To release all resources created here, like disconnecting all queues and the queuemanager*/
	public void releaseResources() 
	{
		String methodName = "releaseResources";
		ld.logData_Debug(" { "+className+"."+methodName);
		try
		{
			if(qopsList != null)
				for(qopsListIndex=0;qopsListIndex<qopsListSize;qopsListIndex++)
				{
					qopsList[qopsListIndex].getDisconnected();
					ld.logData_Debug("  Disconnected  Queue  : "+qopsList[qopsListIndex].queueName);
				}
				if(qMgr != null)
					qMgr.disconnect();
				ld.logData_Debug("  Disconnected QueueManager : "+NodeAttributes.queueManager);
		}
		catch (MQException ex)
		{
			ld.logData_Error("\tWMQ Exception Occurred, Completion code " +ex.completionCode + " Reason code " + ex.reasonCode);
			ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",ex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR",ld.getExceptionString(ex));
		}
		ld.logData_Debug(" } "+className+"."+methodName);
	}
	
	/*Method to get the message from the current indexed queue.....
	Read a message from a queue and pass it to the node....thus read QUEUE_BATCH_SIZE number 
	of messages from a queue before moving to antoher queue...
	if the queue is empty then switch to next queue*/
	public byte[] getMessage() 
	{
		String methodName = "getMessage";
		ld.logData_Debug(" { "+className+"."+methodName);
		byte[] msgText = null;
		QueueOps qops = null;
		try
		{
			
			/*ld.logData_Debug("    Index = "+qopsListIndex);*/
		  	qops = qopsList[qopsListIndex]; /*get the current queue to be read*/
			if(qops.msgCount < qops.QUEUE_BATCH_SIZE) 
			{
				msgText = qops.getMQMessage(); /*get the message*/
				if (msgText != null)
					{
						ld.logData_Debug("  Message received ' "+ msgText +" '");
						qops.msgCount++;
						/*ld.logData_Debug("    Index after = "+qopsListIndex);*/
						return msgText;
					}
				else /*this conditions comes into picture if there are no more msgs 
				to read for the current Queue ..so switch to next queue*/
					{
						qops.msgCount = 0;
						if(qopsListIndex < qopsListSize-1)
							qopsListIndex++;/*increment the current index to the next queue*/
						else qopsListIndex = 0;
					}
			}
			else /*this condition comes into picture if the QUEUE_BATCH_SIZE messages are read
			 for one Queue is over...so switch to next queue*/
			{
				qops.msgCount = 0;
				ld.logData_Debug("  Switching to next queue ");
				if(qopsListIndex < qopsListSize-1)
					qopsListIndex++; /*increment the current index to the next queue*/
				else qopsListIndex = 0;
				/*if(last element then return null) don't switch to begin as it might cause infinite loop*/
			}
	 	}
		catch (IOException ioe)
		{
		  ld.logData_Error("\tUnpredicted Exception Occurred ",ioe);	
		  OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"INTERNAL_UNKNOWN_ERROR", ld.getExceptionString(ioe));
		}
		ld.logData_Debug(" } "+className+"."+methodName);
		/*ld.logData_Debug("    Index after = "+qopsListIndex);*/
		return msgText;
	}
	
	/*To acknowledge the message ..causes the message to be commited on the queue...*/
	public void acknowledgeMessage() 
	{
		String methodName = "acknowledgeMessage";
		ld.logData_Debug(" { "+className+"."+methodName);
		try
		{ 
			qMgr.commit();
			ld.logData_Debug("  Message acknowledged for syncpoint ");
		}
		catch (MQException ex)
		{
			ld.logData_Error("\tWMQ Exception occurred Completion code " +ex.completionCode + " Reason code " + ex.reasonCode);
			ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",ex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR",ld.getExceptionString(ex));
		}
		ld.logData_Debug(" } "+className+"."+methodName);
	}
	
	/*To backout the message ..causes the message to be uncommited on the queue..*/
	public void backoutMessage() throws MbFatalException
	{
		String methodName = "backoutMessage";
		ld.logData_Debug(" { "+className+"."+methodName);
		try
		{ 
			/*check the backout count here and find if the queue has Backout Queue defined,
			 *  if yes, place the message in that queue and commit it.
			 * else find DLQ and place it there.
			 * check current queue retrieved msg bkcnt <= current Q threashold        
			*/
			 if (qopsList[qopsListIndex].msgRcvd.backoutCount <= qopsList[qopsListIndex].backoutThreashold) 
			 {
//			  log here that it is backed out.
       			ld.logData_Debug("  Message backedout, backout count is : "+qopsList[qopsListIndex].msgRcvd.backoutCount);
	            qMgr.backout(); // increase backout count by one 
	         } else 
	         {
            /* Threshold is reached.  Move message to the backout queue */
	            ld.logData_Info("Poisoned message is moving to " + qopsList[qopsListIndex].backoutReQName);
				qopsList[qopsListIndex].putMessageInBackoutQueue(qMgr); //put it in the backout Q
	            qMgr.commit();
	         }
         }
		catch (MQException ex)
		{
			ld.logData_Error("\tWMQ Exception occurred Completion code " +ex.completionCode + " Reason code " + ex.reasonCode);
			ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",ex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR",ld.getExceptionString(ex));
			
			/*handle the error here if the backout message put is failed either because of
			 * the Q not being define or its full. Place it in DLQ.*/
			 try
			 {
		    	if( (ex.reasonCode == MqConstants.MQRC_Q_FULL) || (ex.reasonCode == MqConstants.MQRC_UNKNOWN_OBJECT_NAME) )
		    	 {	
   	 				ld.logData_Error("\t Backout ReQueue Message put failure. Reason code " + ex.reasonCode);
					ld.logData_Error("\t Now Put the message in Qmgr's Dead Letter Queue");
		    	 	qopsList[qopsListIndex].putMessageInDeadLetterQueue(qMgr,deadLetterQueueName,ex.reasonCode); //put it in the DLQ	
		    	 	OSLogData.syslog(className,methodName,OSLogData.SYSLOG_INFO,"WBIMB_DEAD_INFO",ld.getExceptionString(ex));
		    	 	qMgr.commit();
		    	 }
			 }catch(MQException mqex)
			 {
				  /* Terribly you have reached a stage where the poison message cann't be put on DLQ either...*/
		  			ld.logData_Error("\tTerrible error while putting the message on Dead Letter Queue");
	  			  /* This should be handled at customers will in the way its needed.
	  			   * Currently we are going to throw fatal exception.*/
	  			  	ld.logData_Error("\tWMQ Exception occurred Completion code " +mqex.completionCode + " Reason code " + mqex.reasonCode);
					ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",mqex);
					OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"WBIMB_DEAD_ERROR",ld.getExceptionString(mqex));
					throw new MbFatalException(className,methodName,OSLogData.resource,"WBIMB_DEAD_ERROR",ld.getExceptionString(mqex),null);
			 }
			 catch (IOException ioe)
			{
		  		ld.logData_Error("\tTerrible unpredicted Exception Occurred during DLH generation ",ioe);	
		  		OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"INTERNAL_UNKNOWN_ERROR",ld.getExceptionString(ioe));
			}
		}
		ld.logData_Debug(" } "+className+"."+methodName);
	}
	
	/*Internal function to parse the queue(s) and create a list of queue names in an array*/
	private void parseQueueList() 
	{
		String methodName = "parseQueueList";
		ld.logData_Debug(" { "+className+"."+methodName);
		try
		{
			StringTokenizer st = new StringTokenizer(NodeAttributes.queues,",");
			qopsListSize = st.countTokens();
			qopsList = new QueueOps[qopsListSize];
			ld.logData_Debug("  Number of queues found to be : "+qopsListSize);
			for(qopsListIndex=0;qopsListIndex <= qopsListSize;qopsListIndex++)
			{
				qopsList[qopsListIndex] = new QueueOps(st.nextToken(),qMgr);
			}
			qopsListIndex=0;	
		}
		catch(MQException mqex)
		{
			ld.logData_Error("\tWMQ Exception occurred Completion code " +mqex.completionCode + " Reason code " + mqex.reasonCode);
			ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",mqex);
			OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR",ld.getExceptionString(mqex));
		}
		catch(NoSuchElementException  e)
		{
			//System.out.println(e.printStackTrace());
			//OSLogData.syslog(this,"MQOps::parseQueueList",OSLogData.SYSLOG_WARN,"INTERNAL_UNKNOWN_WARN", ld.getExceptionString(e));
		}
		ld.logData_Debug(" } "+className+"."+methodName);
	}

}
