/*****************************************************************************/
/*                                                                           */
/* You may copy, modify, and distribute this sample program in any           */
/* form without payment to IBM,  for any purpose including developing,       */
/* using, marketing or distributing programs that include or are             */
/* derivative works of the sample program.                                   */
/*                                                                           */
/* The sample program is provided to you on an "AS IS" basis, without        */
/* warranty of any kind.  IBM HEREBY EXPRESSLY DISCLAIMS ALL WARRANTIES,     */
/* EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.       */
/* Some jurisdictions do not allow for the exclusion or limitation of        */
/* implied warranties, so the above limitations or exclusions may not        */
/* apply to you.  IBM shall not be liable for any damages you suffer as      */
/* a result of using, modifying or distributing the sample program or        */
/* its derivatives.                                                          */
/*																			 */
/*  * @author Shravan														 */
/*                                                                           */
/*                                                                           */
/*This is the core/main file of MQ Queue side operations.                    */
/* V 1.0 - Basic features						     						 */
/* V 1.4 - Feature - Support for OS Logging					     			 */
/* V 1.7 - Bug fix - messages should be under syncpoint						 */
/* V 1.8 - Feature - New property added to handle messages under syncpoint.  */
/* 			This is similar to buil-in input node, where messages transaction*/
/* 			processing will be done depending on the node attributes 		 */
/* 			'transaction' value. Three new functions 'generateMQDLH'		 */
/* 			'putMessageInBackoutQueue' and 'putMessageInDeadLetterQueue' are */
/*			added to handle this feature. Appropriate modifications are done */
/* 			in MQOps.java and AnyQmgrInputNode.java							 */
/* V 2.0 - Change wrt OsLogData first parameter								 */
/*****************************************************************************/
package com.ibm.Shra.Xternal.Plugins;

import com.ibm.mq.MQQueue;
import com.ibm.mq.MQQueueManager;
import com.ibm.mq.MQMessage;
import com.ibm.mq.MQGetMessageOptions;
import com.ibm.mq.MQPutMessageOptions;
import com.ibm.mq.MQException;
import com.ibm.mq.MQC;
import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.util.GregorianCalendar;
import java.util.Calendar;

class QueueOps
{
	private String className = "queueOps";
	private static logData ld =  null;
	
	private MQQueue lq = null;/*Object to point to the current MQ Queue*/
/*	private MQQueueManager qMgr = null;/*Object to point to the current MQ Queue Manager...we don't need to maintain it.*/
	MQMessage msgRcvd = null; /*Object to point to the current MQ Message*/
	private MQGetMessageOptions gmo = null; /*Object to point to the current MQ Get Message Options*/
	String queueName = null; /*String to store the name of queue for debug purpose*/
	int msgCount=0; /*Index to count how many messages are read from a queue*/
	int QUEUE_BATCH_SIZE = 5; /*Index to indicate how many messages to be read maximum from a queue 
	before moving o other queue*/
	
	private String osVersion = "Windows 2k"; /*we will check only for the word "Window" used to handle endian problem*/
	private boolean osWin = false; /*set this true if os is windows so that our MQMD header is little-endian for
	other platforms it is big-endian*/
/*			  Declarations for transaction support */	
	int backoutThreashold;	 
	String backoutReQName = null;
	
	/*Initialize the basic queue options*/
	public QueueOps(String qname, MQQueueManager qMgr) throws MQException
	{
		ld = logData.getInstance();
		ld.logData_Debug(" { "+className);	
		if(osVersion.regionMatches(true,0,System.getProperty("os.name"),0,5))
			{
				ld.logData_Debug("  OS found to be windows version ");
				osWin = true; /*We got to do this to convert the type 'int' in 
				header into little endian for the sake of broker*/
			}
		else 
			{	ld.logData_Debug("  OS found to be NOT windows version ");
			    osWin = false;
			}
		queueName = qname;
		ld.logData_Debug("  Queue name set to : "+qname);
		/* Open the queue for input and inquire.  
		 The inquire is needed so that the program can get the backout queue and threshold.*/
		 int[]  selectors = { MqConstants.MQIA_BACKOUT_THRESHOLD, MqConstants.MQCA_BACKOUT_REQ_Q_NAME	};
		 int[]  intAttrs = new int[1];
		 byte[] charAttrs = new byte[MqConstants.MQ_Q_NAME_LENGTH];
		 /*first open the queue for querying and then close it.*/
	     int openOptions = MQC.MQOO_INPUT_AS_Q_DEF | MQC.MQOO_INQUIRE | MQC.MQOO_BROWSE | MQC.MQOO_SAVE_ALL_CONTEXT;
	      lq = qMgr.accessQueue(qname, openOptions, null, null, null);
	      lq.inquire(selectors, intAttrs, charAttrs);
     	  backoutThreashold = intAttrs[0]; /*set the backout threashold, if incase the message processing fails
     	  		then the message is backed out on the threashold value.*/
   	      backoutReQName = new String(charAttrs);/*set the backout queue name.*/
			
	   	  ld.logData_Debug("  Backout Threashold of the queue "+backoutThreashold);
  	   	  ld.logData_Debug("  Backout ReQueue name is "+backoutReQName);
  	   	  getDisconnected(); //Close the connection as u don't need it for more inquire.
		msgRcvd = new MQMessage();
		ld.logData_Debug("  Message object created ");
		gmo = new MQGetMessageOptions(); 
		gmo.options = gmo.options + MQC.MQGMO_NO_WAIT + MQC.MQGMO_FAIL_IF_QUIESCING; /*Change this to your needs.*/
		/*Decide on the transaction that customer has chosen and set the options accordingly*/
		if(NodeAttributes.transaction.equals("Yes"))
			gmo.options += MQC.MQGMO_SYNCPOINT;
		else if(NodeAttributes.transaction.equals("No"))
			gmo.options += MQC.MQGMO_NO_SYNCPOINT;
		else if(NodeAttributes.transaction.equals("Automatic"))
			gmo.options += MQC.MQGMO_SYNCPOINT_IF_PERSISTENT;
			
		ld.logData_Debug("  Get Message Options set ");
		ld.logData_Debug(" } "+className);	
	}
	/*Connect to the queue*/
	public void getConnected(MQQueueManager qMgr) throws MQException
	{
		ld.logData_Debug(" { "+className+".getConnected ");
		int openOptions = MQC.MQOO_INPUT_AS_Q_DEF + MQC.MQOO_FAIL_IF_QUIESCING;/*Change this to your needs*/
		lq = qMgr.accessQueue(queueName,openOptions);
		ld.logData_Debug("  Connected to Queue : "+ queueName);
		ld.logData_Debug(" } "+className+".getConnected ");
	}
	/*Function to read the MQMessage from the Queue*/
	public byte[] getMQMessage() throws IOException
	{
		String methodName = "getMQMessage";
		byte[] msgText = null;
		ld.logData_Debug(" { "+className+"."+methodName);
		try
		{
			msgRcvd.clearMessage();
		        msgRcvd.correlationId = MQC.MQCI_NONE;
		        msgRcvd.messageId     = MQC.MQMI_NONE;
			lq.get(msgRcvd, gmo);

			int msgLen = msgRcvd.getTotalMessageLength();/*get the length of the message*/
			byte strData[] = new byte[msgLen];
			msgRcvd.readFully(strData,0,msgLen);/*read in bytes*/
			ld.logData_Debug("  Message received from queue : "+queueName+" ' "+ new String(strData) +" '");
			byte strMQMD[] = generateMQMD(msgRcvd); /*generate MQMD header*/
			int length = strData.length + strMQMD.length;/*gives total length of message along with header*/
			msgText = new byte[length];
			int i=0,j=0;
			for(; i < strMQMD.length; i++)
			  msgText[i] = strMQMD[i]; /*add the header first to the outgoing message*/
			for(; j < strData.length; i++,j++)
			  msgText[i] = strData[j];/*add the message body received first to the outgoing message*/
			ld.logData_Debug("  Message generated ' "+ new String(msgText) +" '");
	 	}catch (MQException ex)
		{
			if (ex.reasonCode == MqConstants.MQRC_NO_MSG_AVAILABLE)
			{
			  ld.logData_Debug("\tIgnore Exception as there are no messages in the Queue ; "+queueName+" , switch to next queue");
			}
			else if ( (ex.reasonCode == MqConstants.MQRC_CONNECTION_BROKEN) || 
					(ex.reasonCode == MqConstants.MQRC_CONNECTION_QUIESCING ) ||
					(ex.reasonCode == MqConstants.MQRC_Q_MGR_QUIESCING ) ||
					(ex.reasonCode == MqConstants.MQRC_Q_MGR_STOPPING) )
			{
				ld.logData_Error("\t Fatal error. Queue Manager not available. WMQ Exception Occurred, Completion code " +ex.completionCode + " Reason code " + ex.reasonCode);
				ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",ex);
				OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR", ld.getExceptionString(ex));
			}
			else
			{
				ld.logData_Error("\t Failed to read message. WMQ Exception Occurred, Completion code " +ex.completionCode + " Reason code " + ex.reasonCode);
				ld.logData_Error("\tWMQ Exception Detailed Stack Trace : ",ex);
				OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"MQ_ERROR",ld.getExceptionString(ex));
			}
		}
		 ld.logData_Debug(" } "+className+"."+methodName);
		return msgText;			 
	}
	/*Disconnect the Queue connection*/
	public void getDisconnected() throws MQException
	{
		lq.close();	
	}
	/*Generate MQMD header in bytes...if windows it should be little-endian else it got to be big-endian */
	private byte[] generateMQMD(MQMessage mqmsg)
	{
		String methodName = "generateMQMD";
		ld.logData_Debug("  { "+className+"."+methodName);			
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		DataOutputStream dataBuffer = new DataOutputStream(baos);
	  try
	   {
	    dataBuffer.writeBytes("MD  ");													//StrucId
	    
	    //ld.logData_Debug("   vesrion  "+mqmsg.getVersion());							//Version
	    if (osWin)																		
			 	dataBuffer.writeInt(reverseInt(mqmsg.getVersion()));
		else 	dataBuffer.writeInt(mqmsg.getVersion());
		
		//ld.logData_Debug("   report  "+mqmsg.report);									//Report
		if (osWin)												
				dataBuffer.writeInt(reverseInt(mqmsg.report));
    	else 	dataBuffer.writeInt(mqmsg.report);									
    	
    	//ld.logData_Debug("   messageType  "+mqmsg.messageType);						//MsgType
    	if (osWin)												
    	    	dataBuffer.writeInt(reverseInt(mqmsg.messageType));
    	else 	dataBuffer.writeInt(mqmsg.messageType);
    	
    	//ld.logData_Debug("   expiry  "+mqmsg.expiry);									//Expiry
    	if (osWin)												
    			dataBuffer.writeInt(reverseInt(mqmsg.expiry));
    	else 	dataBuffer.writeInt(mqmsg.expiry);
    	
    	//ld.logData_Debug("   feedback  "+mqmsg.feedback);								//Feedback
    	if (osWin)													
    			dataBuffer.writeInt(reverseInt(mqmsg.feedback));
    	else 	dataBuffer.writeInt(mqmsg.feedback);
    	
    	//ld.logData_Debug("   encoding  "+mqmsg.encoding);								//Encoding
    	if (osWin)													
    			dataBuffer.writeInt(reverseInt(mqmsg.encoding));
    	else	dataBuffer.writeInt(mqmsg.encoding);
    	
    	//ld.logData_Debug("   characterSet  "+mqmsg.characterSet);						//CodedCharSetId
		if (osWin)															
				dataBuffer.writeInt(reverseInt(mqmsg.characterSet));
    	else 	dataBuffer.writeInt(mqmsg.characterSet);
    	
    	//ld.logData_Debug("   format  "+mqmsg.format);									//Format
    	dataBuffer.writeBytes(setStringToSize(mqmsg.format,8));			
    	
    	//ld.logData_Debug("   priority  "+mqmsg.priority);								//Priority
    	if (osWin)												
    			dataBuffer.writeInt(reverseInt(mqmsg.priority));
    	else 	dataBuffer.writeInt(mqmsg.priority);
    		
    	//ld.logData_Debug("   persistence  "+mqmsg.persistence);						//Persistence
    	if (osWin)															
    	   		dataBuffer.writeInt(reverseInt(mqmsg.persistence));
    	else	dataBuffer.writeInt(mqmsg.persistence);
/*		if(NodeAttributes.transaction.equals("Automatic"))
			if(mqmsg.persistence == MQC.MQPER_PERSISTENT)
			  NodeAttributes.transaction = "Yes";
*/    	
    	//ld.logData_Debug("   messageId  "+mqmsg.messageId);							//MsgId
   		dataBuffer.write(mqmsg.messageId,0,24 );						
   		
   		//ld.logData_Debug("   correlationId  "+mqmsg.correlationId);					//CorrelId
   		dataBuffer.write(mqmsg.correlationId,0,24);								
   		
	 	//ld.logData_Debug("   backoutCount  "+mqmsg.backoutCount);						//BackoutCount
	 	if (osWin)															
	 			dataBuffer.writeInt(reverseInt(mqmsg.backoutCount)); 							
	 	else 	dataBuffer.writeInt(mqmsg.backoutCount);
	 	
	 	//ld.logData_Debug("   replyToQueueName  "+mqmsg.replyToQueueName);				//ReplyToQ
   		dataBuffer.writeBytes(setStringToSize(mqmsg.replyToQueueName,48));		
   		
   		//ld.logData_Debug("   replyToQueueManagerName  "+mqmsg.replyToQueueManagerName);//ReplyToQMgr
   		dataBuffer.writeBytes(setStringToSize(mqmsg.replyToQueueManagerName,48));		
   			
   		//ld.logData_Debug("   userId  "+mqmsg.userId);									//UserIdentifier
		dataBuffer.writeBytes(setStringToSize(mqmsg.userId,12));
		
		//ld.logData_Debug("   accountingToken  "+mqmsg.accountingToken);				//Accounting Token
   		dataBuffer.write(mqmsg.accountingToken,0,32);								
   		
   		//ld.logData_Debug("   applicationIdData  "+mqmsg.applicationIdData);			//ApplIdentityData
   		dataBuffer.writeBytes(setStringToSize(mqmsg.applicationIdData,32));			
   		
   		
    	//ld.logData_Debug("   putApplicationType  "+mqmsg.putApplicationType);			//PutApplType
    	if (osWin)																		
    			dataBuffer.writeInt(reverseInt(mqmsg.putApplicationType));
    	else	dataBuffer.writeInt(mqmsg.putApplicationType);
    	
    	//ld.logData_Debug("   putApplicationName  "+mqmsg.putApplicationName);			//PutApplName
   		dataBuffer.writeBytes(setStringToSize(mqmsg.putApplicationName,28));			
   		
   		
   		//ld.logData_Debug("   putDateTime  ");											//PutDateTime
   		dataBuffer.writeBytes(setStringToSize(getDateTimeString(mqmsg.putDateTime),16));
   		
   		//ld.logData_Debug("   applicationOriginData  "+mqmsg.applicationOriginData);	//ApplOriginData
    	dataBuffer.writeBytes(setStringToSize(mqmsg.applicationOriginData,4));	
    	
    	//ld.logData_Debug("   groupId  "+mqmsg.groupId);								//GroupId    	
    	dataBuffer.write(mqmsg.groupId,0,24);						

    	//ld.logData_Debug("   messageSequenceNumber  "+mqmsg.messageSequenceNumber);	//MsgSeqNumber
      	if (osWin)															
      			dataBuffer.writeInt(reverseInt(mqmsg.messageSequenceNumber));
      	else 	dataBuffer.writeInt(mqmsg.messageSequenceNumber);

      	//ld.logData_Debug("   offset  "+mqmsg.offset);									//Offset      	
      	if (osWin)												
      			dataBuffer.writeInt(reverseInt(mqmsg.offset)); 									
      	else 	dataBuffer.writeInt(mqmsg.offset);

      	
      	//ld.logData_Debug("   messageFlags  "+mqmsg.messageFlags);						//MsgFlags
      	if (osWin)															
      			dataBuffer.writeInt(reverseInt(mqmsg.messageFlags)); 							
      	else 	dataBuffer.writeInt(mqmsg.messageFlags);
      	
      	//ld.logData_Debug("   originalLength  "+mqmsg.originalLength);					//OriginalLength
      	if (osWin)																
      			dataBuffer.writeInt(reverseInt(mqmsg.originalLength));
      	else	dataBuffer.writeInt(mqmsg.originalLength);
      	
      	
    	baos.close();
       }catch(Exception e)
	  {
  		  ld.logData_Error("\tWMQ Exception Occurred while building the MQMD header ",e);
		  OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"INTERNAL_UNKNOWN_ERROR",ld.getExceptionString(e));	  
	  }
       
       ld.logData_Debug("   Required Header Generated ");
       ld.logData_Debug("  } "+className+"."+methodName);			
		return baos.toByteArray();
	}
	/*This is not useful..but better not take a chance as we are building the bytes, lets set the string length
	to specific size of the element in the Header...used only by generateMQMD() function*/
	private String setStringToSize( String s, int length ) 
	{     
       String retVal = s ;
	   if ( s == null ) /*Am aware....this situations is rare but still.....*/
	   {
	      retVal = "" ;
	      for ( int i = 0 ; i < length ; i++ ) 
      		  retVal += " " ;
       } 
       else 
       {  
      	for ( int i = s.length( ); i < length ; i++ ) 
          retVal = retVal + " " ;
       }
       return retVal;
    }           
    /*We got to have this function only for windows as we got to convert the Java big-endian 'int'
    type to little-endian in the MQMD Header for the sake of broker who lives on C/C++ */
	private int reverseInt(int v) 
	{
     return  (v >>> 24) | (v << 24) | ((v << 8) & 0x00FF0000) | ((v >> 8) & 0x0000FF00);
    }
    /*What you get from MQMessage is a GregorianCalendar and we got to convert this to bytes/String
    while constructing the MQMD header*/
    private String getDateTimeString(GregorianCalendar gc)
	{
		StringBuffer sb = new StringBuffer();
		sb.append(gc.get(Calendar.YEAR));//year 4chars
		
		int month = gc.get(Calendar.MONTH) + 1; //month 2 chars
		if (month < 10) 
		   sb.append('0');
		sb.append(month);
      	
        int day = gc.get(Calendar.DAY_OF_MONTH); //day 2 chars
        if (day < 10) 
           sb.append('0');
		sb.append(day);
      	
        int hour = gc.get(Calendar.HOUR_OF_DAY);//hour 2 chars
        if (hour < 10) 
           sb.append('0');
        sb.append(hour);
      	
        int minutes = gc.get(Calendar.MINUTE);//minutes 2 chars
        if (minutes < 10) 
           sb.append('0');
        sb.append(minutes);
      	
        int seconds = gc.get(Calendar.SECOND); //seconds 2 chars
        if (seconds < 10)
           sb.append('0');
        sb.append(seconds);
      	
        int millisecs = gc.get(Calendar.MILLISECOND);//milli seconds 2 chars
        millisecs = millisecs/10;
        if (millisecs < 10) 
           sb.append('0');
        sb.append(millisecs);
       	      	
        return (new String(sb.toString()));
	}
	/* Once the message processing failed for threashold number of times then put the message in backout queue.*/
	void putMessageInBackoutQueue(MQQueueManager qMgr) throws MQException
	{
 		String methodName = "putMessageInBackoutQueue";
		ld.logData_Debug("  { "+className+"."+methodName);			
		/*Set the ppen queue options*/
		 int openPutOptions = MQC.MQOO_OUTPUT + MQC.MQOO_FAIL_IF_QUIESCING;
	     MQQueue myPutQueue = qMgr.accessQueue(backoutReQName, openPutOptions,null, null, null);
         MQPutMessageOptions pmo = new MQPutMessageOptions();
         /*Decide the transaction type that customer has chosen.*/
  		 if(NodeAttributes.transaction.equals("Yes"))
			pmo.options += MQC.MQGMO_SYNCPOINT;
		 else if(NodeAttributes.transaction.equals("No"))
			pmo.options += MQC.MQGMO_NO_SYNCPOINT;
		 else if(NodeAttributes.transaction.equals("Automatic"))
			pmo.options += MQC.MQGMO_SYNCPOINT_IF_PERSISTENT;
         pmo.contextReference = lq;
         myPutQueue.put(msgRcvd, pmo);
         myPutQueue.close();
        ld.logData_Info("Message put in Backout Queue Successful");
   		ld.logData_Debug("  } "+className+"."+methodName);			
	}
/*
	 * You will reach here only if the message processing is failed before reaching the MQOutPutNode.
	 * So, we have to put the message in backout queue, if that fails then we have to put the message in 
	 * DeadLetterQueue.
*/
	void putMessageInDeadLetterQueue(MQQueueManager qMgr, String deadLetterQueueName,int reasonCode) throws MQException,IOException
	{
 		String methodName = "putMessageInDeadLetterQueue";
		ld.logData_Debug("  { "+className+"."+methodName);			

		/* we have the message still in our hands which got backedout/ failed.
		 * so first modify the MQMD header appropriately for the DLQ.
		 */
		 msgRcvd.format = MQC.MQFMT_DEAD_LETTER_HEADER;
		 /*Genereally we need to take care of encoding and CCSID, but as the DLQ is on the 
		  * same QMGR as that of the i/p queue, so am not setting them explicitly.*/
		  /*Before we modify the exisiting message with the MQDLH, we shall store the mesagebody.*/
		msgRcvd.seek(0);
		int msgLen = msgRcvd.getTotalMessageLength();/*get the length of the message*/
		byte strData[] = new byte[msgLen];
		msgRcvd.readFully(strData,0,msgLen);/*read in bytes*/
	  
		/*Now change the message such that we have MQDLH between MQMD and message body*/
		msgRcvd.seek(0); /*this should brings us to the tip of the message and tail of the MQMD header.*/
		byte strMQDLH[] = generateMQDLH(msgRcvd,reasonCode);
		msgRcvd.write(strMQDLH); /*Add the DLH header after MQMD*/
		/*The general expected structure is MQMD + MQDLH + Msg Body.*/
		/*Now add the message body*/
		msgRcvd.write(strData);
		strData = null;
		/*built the message with DLH successfully and if put fails then declare backout and declare fatal error*/
	 	int openPutOptions = MQC.MQOO_OUTPUT + MQC.MQOO_FAIL_IF_QUIESCING;
	     MQQueue myPutQueue = qMgr.accessQueue(deadLetterQueueName, openPutOptions,null, null, null);
         MQPutMessageOptions pmo = new MQPutMessageOptions();
  		 if(NodeAttributes.transaction.equals("Yes"))
			pmo.options += MQC.MQGMO_SYNCPOINT;
		 else if(NodeAttributes.transaction.equals("No"))
			pmo.options += MQC.MQGMO_NO_SYNCPOINT;
		 else if(NodeAttributes.transaction.equals("Automatic"))
			pmo.options += MQC.MQGMO_SYNCPOINT_IF_PERSISTENT;
         pmo.contextReference = lq;
         myPutQueue.put(msgRcvd, pmo);
         myPutQueue.close();
         ld.logData_Info("Message put in Dead Letter Queue Successful");
   		ld.logData_Debug("  } "+className+"."+methodName);			
	}
	/*Generate MQDLH header in bytes...if windows it should be little-endian else it got to be big-endian 
	 */
	private byte[] generateMQDLH(MQMessage mqmsg,int reasonCode)
	{
		String methodName = "generateMQDLH";
		ld.logData_Debug("  { "+className+"."+methodName);			
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		DataOutputStream dataBuffer = new DataOutputStream(baos);
	  try
	   {
	    dataBuffer.writeBytes("DLH ");													//StrucId
	    
	    //ld.logData_Debug("   vesrion  "+mqmsg.getVersion());							//Version
	    if (osWin)																		
			 	dataBuffer.writeInt(reverseInt(1));
		else 	dataBuffer.writeInt(1);
		
		//ld.logData_Debug("   reason  ");												//Reason
		if (osWin)												
				dataBuffer.writeInt(reverseInt(reasonCode));
    	else 	dataBuffer.writeInt(reasonCode);									

	 	//*ld.logData_Debug("   replyToQueueName  "+mqmsg.replyToQueueName);				//ReplyToQ
   		dataBuffer.writeBytes(setStringToSize(this.backoutReQName,48));		
   		
   		//*ld.logData_Debug("   replyToQueueManagerName  "+mqmsg.replyToQueueManagerName);//ReplyToQMgr
   		dataBuffer.writeBytes(setStringToSize(NodeAttributes.queueManager,48));		
    	
    	//*ld.logData_Debug("   encoding  "+mqmsg.encoding);								//Encoding
    	if (osWin)													
    			dataBuffer.writeInt(reverseInt(mqmsg.encoding));
    	else	dataBuffer.writeInt(mqmsg.encoding);
    	
    	//*ld.logData_Debug("   characterSet  "+mqmsg.characterSet);						//CodedCharSetId
		if (osWin)															
				dataBuffer.writeInt(reverseInt(mqmsg.characterSet));
    	else 	dataBuffer.writeInt(mqmsg.characterSet);
    	
    	//*ld.logData_Debug("   format  "+mqmsg.format);									//Format
    	dataBuffer.writeBytes(setStringToSize(mqmsg.format,8));			
    	
    	//*ld.logData_Debug("   putApplicationType  "+mqmsg.putApplicationType);			//PutApplType
    	if (osWin)																		
    			dataBuffer.writeInt(reverseInt(MQC.MQAT_DEFAULT));
    	else	dataBuffer.writeInt(mqmsg.putApplicationType);
    	
    	//*ld.logData_Debug("   putApplicationName  "+mqmsg.putApplicationName);			//PutApplName
   		//The same is set when the built-in input node is used.		
// 		dataBuffer.writeBytes(setStringToSize("MQSeriesIntegrator2",28));	
   		dataBuffer.writeBytes(setStringToSize("AnyQMgrInputNode   ",28));	
   		
   		//*ld.logData_Debug("   putDateTime  ");											//PutDateTime
   		dataBuffer.writeBytes(setStringToSize(getDateTimeString(new GregorianCalendar()),16));
      	
    	baos.close();
       }catch(Exception e)
	  {
  		  ld.logData_Error("\tWMQ Exception Occurred while building the MQDLH header ",e);
		  OSLogData.syslog(className,methodName,OSLogData.SYSLOG_ERROR,"INTERNAL_UNKNOWN_ERROR",ld.getExceptionString(e));	  
	  }
       
       ld.logData_Debug("   Required Dead Letter Header Generated ");
       ld.logData_Debug("  } "+className+"."+methodName);			
		return baos.toByteArray();
	}

}
